import json
import os

from pokemongo_bot import inventory
from pokemongo_bot.base_dir import _base_dir
from pokemongo_bot.base_task import BaseTask
from pokemongo_bot.human_behaviour import action_delay
from pokemongo_bot.services.item_recycle_worker import ItemRecycler
from pokemongo_bot.tree_config_builder import ConfigException
from pokemongo_bot.worker_result import WorkerResult
from random import uniform
from datetime import datetime as dt, timedelta


DEFAULT_MIN_EMPTY_SPACE = 6

class RecycleItems(BaseTask):
    """
    Recycle undesired items if there is less than five space in inventory.
    You can use either item's name or id. For the full list of items see ../../data/items.json
    
    Can also force a recycle to occur at a pseudo-random time between recycle_force_min and 
    recycle_force_max minutes.

    It's highly recommended to put this task before move_to_fort and spin_fort task in the config file so you'll most likely be able to loot.

    Example config :
    {
      "type": "RecycleItems",
      "config": {
        "min_empty_space": 6,           # 6 by default
        "max_balls_keep": 150,
        "max_potions_keep": 50,
        "max_berries_keep": 70,
        "max_revives_keep": 70,
        "item_filter": {
          "Pokeball": {"keep": 20},
          "Greatball": {"keep": 50},
          "Ultraball": {"keep": 100},
          "Potion": {"keep": 0},
          "Super Potion": {"keep": 0},
          "Hyper Potion": {"keep": 20},
          "Max Potion": {"keep": 50},
          "Revive": {"keep": 0},
          "Max Revive": {"keep": 20},
          "Razz Berry": {"keep": 20}
        },
        "recycle_wait_min": 1,
        "recycle_wait_max": 4,
        "recycle_force": true,
        "recycle_force_min": "00:00:00",
        "recycle_force_max": "00:01:00"
      }
      
    }
    """
    SUPPORTED_TASK_API_VERSION = 1


    def initialize(self):
        self.items_filter = self.config.get('item_filter', {})
        self.min_empty_space = self.config.get('min_empty_space', None)
        self.max_balls_keep = self.config.get('max_balls_keep', None)
        self.max_potions_keep = self.config.get('max_potions_keep', None)
        self.max_berries_keep = self.config.get('max_berries_keep', None)
        self.max_revives_keep = self.config.get('max_revives_keep', None)
        self.recycle_wait_min = self.config.get('recycle_wait_min', 1)
        self.recycle_wait_max = self.config.get('recycle_wait_max', 4)
        self.recycle_force = self.config.get('recycle_force', False)
        self.recycle_force_min = self.config.get('recycle_force_min', '00:01:00')
        self.recycle_force_max = self.config.get('recycle_force_max', '00:10:00')       
        self.minInterval = self.getSeconds(self.recycle_force_min)
        self.maxInterval = self.getSeconds(self.recycle_force_max)
        self._validate_item_filter()
        
        if self.recycle_force:
            self._schedule_next_force()

    def getSeconds(self, strTime):
        '''
        Return the duration in seconds of a time string
        :param strTime: string time of format %H:%M:%S
        '''
        try:
            x = dt.strptime(strTime, '%H:%M:%S')
            seconds = int(timedelta(hours=x.hour,minutes=x.minute,seconds=x.second).total_seconds())
        except ValueError: 
            seconds = 0;
            
        if seconds < 0:
            seconds = 0;
              
        return seconds

    def _schedule_next_force(self):
        '''
        Schedule the time aof the next forced recycle.
        '''
        self._next_force = self._get_next_force_schedule()
        self.emit_event(
            'next_force_recycle',
            formatted="Next forced item recycle at {time}",
            data={
                'time': str(self._next_force.strftime("%H:%M:%S"))
            }
        )

    def _should_force_now(self):
        if dt.now() >= self._next_force:
            return True

        return False

    def _get_next_force_schedule(self):
        now = dt.now()
        next_time = now + timedelta(seconds=int(uniform(self.minInterval, self.maxInterval)))

        return next_time

    def _validate_item_filter(self):
        """
        Validate user's item filter config
        :return: Nothing.
        :rtype: None
        :raise: ConfigException: When an item doesn't exist in ../../data/items.json
        """
        item_list = json.load(open(os.path.join(_base_dir, 'data', 'items.json')))
        for config_item_name, bag_count in self.items_filter.iteritems():
            if config_item_name not in item_list.viewvalues():
                if config_item_name not in item_list:
                    raise ConfigException(
                        "item {} does not exist, spelling mistake? (check for valid item names in data/items.json)".format(
                            config_item_name))

    def should_run(self):
        """
        Returns a value indicating whether the recycling process should be run.
        :return: True if the recycling process should be run; otherwise, False.
        :rtype: bool
        """
        
        if self.recycle_force and self._should_force_now():
            self.emit_event(
                'force_recycle',
                formatted="Forcing item recycle based on schedule"
            )
            self._schedule_next_force()
            return True
        
        if inventory.Items.get_space_left() <= (DEFAULT_MIN_EMPTY_SPACE if self.min_empty_space is None else self.min_empty_space):
            return True
        return False

    def work(self):
        """
        Start the process of recycling items if necessary.
        :return: Returns whether or not the task went well
        :rtype: WorkerResult
        """

        worker_result = WorkerResult.SUCCESS
        if self.should_run():

            if not (self.max_balls_keep is None):
                this_worker_result = self.recycle_excess_category_max(self.max_balls_keep, [1,2,3,4])
                if this_worker_result != WorkerResult.SUCCESS:
                    worker_result = this_worker_result
                    
            if not (self.max_potions_keep is None):
                this_worker_result =  self.recycle_excess_category_max(self.max_potions_keep, [101,102,103,104])
                if this_worker_result != WorkerResult.SUCCESS:
                    worker_result = this_worker_result
                   
            if not (self.max_berries_keep is None):
                this_worker_result =  self.recycle_excess_category_max(self.max_berries_keep, [701,702,703,704,705])
                if this_worker_result != WorkerResult.SUCCESS:
                    worker_result = this_worker_result
                    
            if not (self.max_revives_keep is None):
                this_worker_result = self.recycle_excess_category_max(self.max_revives_keep, [201,202])
                if this_worker_result != WorkerResult.SUCCESS:
                    worker_result = this_worker_result
                    
            for item_in_inventory in inventory.items().all():
                if self.item_should_be_recycled(item_in_inventory):
                    # Make the bot appears more human
                    action_delay(self.recycle_wait_min, self.recycle_wait_max)
                    # If at any recycling process call we got an error, we consider that the result of this task is error too.
                    if item_in_inventory.recycle(self.get_amount_to_recycle(item_in_inventory)) == WorkerResult.ERROR:
                        worker_result = WorkerResult.ERROR
           
        return worker_result

    def recycle_excess_category_max(self, category_max, category_items_list):
        """
        Recycle the item which excess the category max
        :param category_max:
        :param category_items_list:
        :return: none:
        :rtype: None
        """
        worker_result = WorkerResult.SUCCESS
        category_inventory = self.get_category_inventory_list(category_items_list)
        category_count = 0
        for i in category_inventory:
           category_count = category_count + i[1]
        items_to_recycle = self.get_category_items_to_recycle(category_inventory, category_count, category_max)
        for item in items_to_recycle:
            action_delay(self.recycle_wait_min, self.recycle_wait_max)
            if inventory.items().get(item[0]).recycle(item[1]) == WorkerResult.ERROR:
                worker_result = WorkerResult.ERROR
        return worker_result

    def get_category_inventory_list(self, category_inventory):
        """
        Returns an array of items with the item id and item count.
        :param category_inventory:
        :return: array of items within a category:
        :rtype: array
        """
        x = 0
        category_inventory_list = []
        for c in category_inventory:
            category_inventory_list.append([])
            category_inventory_list[x].append(c)
            category_inventory_list[x].append(inventory.items().get(c).count)
            x = x + 1
        return category_inventory_list

    def get_category_items_to_recycle(self, category_inventory, category_count, category_max):
        """
        Returns an array to be recycle within a category of items with the item id and item count.
        :param category_inventory:
        :param category_count:
        :param category_max:
        :return: array of items to be recycle.
        :rtype: array
        """
        x = 0
        items_to_recycle = []
        if category_count > category_max:
            items_to_be_recycled = category_count - category_max

            for item in category_inventory:
                if items_to_be_recycled == 0:
                    break
                if items_to_be_recycled >= item[1]:
                    items_to_recycle.append([])
                    items_to_recycle[x].append(item[0])
                    items_to_recycle[x].append(item[1])
                else:
                    items_to_recycle.append([])
                    items_to_recycle[x].append(item[0])
                    items_to_recycle[x].append(items_to_be_recycled)
                items_to_be_recycled = items_to_be_recycled - items_to_recycle[x][1]
                x = x + 1
        return items_to_recycle

    def item_should_be_recycled(self, item):
        """
        Returns a value indicating whether the item should be recycled.
        :param item: The Item to test
        :return: True if the title should be recycled; otherwise, False.
        :rtype: bool
        """
        return (item.name in self.items_filter or str(item.id) in self.items_filter) and self.get_amount_to_recycle(item) > 0

    def get_amount_to_recycle(self, item):
        """
        Determine the amount to recycle accordingly to user config
        :param item: Item to determine the amount to recycle.
        :return: The amount to recycle
        :rtype: int
        """
        amount_to_keep = self.get_amount_to_keep(item)
        return 0 if amount_to_keep is None else item.count - amount_to_keep

    def get_amount_to_keep(self, item):
        """
        Determine item's amount to keep in inventory.
        :param item:
        :return: Item's amount to keep in inventory.
        :rtype: int
        """
        item_filter_config = self.items_filter.get(item.name, 0)
        if item_filter_config is not 0:
            return item_filter_config.get('keep', 20)
        else:
            item_filter_config = self.items_filter.get(str(item.id), 0)
            if item_filter_config is not 0:
                return item_filter_config.get('keep', 20)
